""" Tests for exerciser module
"""

from os.path import dirname, join as pjoin
from textwrap import dedent

from rnbgrader import load, loads
from rnbgrader.nbparser import RNotebook

from rmdex.exerciser import (make_check_exercise, make_exercise, make_solution,
                             get_marks, check_marks, check_chunk_marks,
                             question_chunks, MARK_RE, template2exercise,
                             MarkupError, read_utf8)

import pytest


HERE = dirname(__file__)
SOLUTION_FNAME = pjoin(HERE, 'solution.Rmd')
EXERCISE_FNAME = pjoin(HERE, 'exercise.Rmd')
SOLUTION_STR = read_utf8(SOLUTION_FNAME)
EXERCISE_STR = read_utf8(EXERCISE_FNAME)

TEMPLATE_FNAME = pjoin(HERE, 'template.Rmd')
T_EXERCISE_FNAME = pjoin(HERE, 'template_exercise.Rmd')
T_SOLUTION_FNAME = pjoin(HERE, 'template_solution.Rmd')
TEMPLATE_STR = read_utf8(TEMPLATE_FNAME)
T_EXERCISE_STR = read_utf8(T_EXERCISE_FNAME)
T_SOLUTION_STR = read_utf8(T_SOLUTION_FNAME)

FMFT_FNAME = pjoin(HERE, 'fix_my_fors_template.Rmd')
FMFE_FNAME = pjoin(HERE, 'fix_my_fors.Rmd')
FMFS_FNAME = pjoin(HERE, 'fix_my_fors_solution.Rmd')
FMFT_STR = read_utf8(FMFT_FNAME)
FMFE_STR = read_utf8(FMFE_FNAME)
FMFS_STR = read_utf8(FMFS_FNAME)

def test_make_check_exercise():
    assert make_check_exercise(SOLUTION_STR) == EXERCISE_STR


def test_make_exercise():
    nb = load(SOLUTION_FNAME)
    check_marks(nb.nb_str)
    exercise = make_exercise(SOLUTION_STR)
    assert exercise == EXERCISE_STR
    check_marks(exercise)
    check_chunk_marks(question_chunks(loads(exercise)))
    nb = load(TEMPLATE_FNAME)
    exercise = make_exercise(TEMPLATE_STR)
    assert exercise == T_EXERCISE_STR
    nb = load(FMFT_FNAME)
    exercise = make_exercise(FMFT_STR)
    assert exercise == FMFE_STR


def test_make_solution():
    # No changes for the basic example (no #<- lines)
    solution = make_solution(SOLUTION_STR)
    assert solution == SOLUTION_STR
    # The Python example does have #<- lines.
    solution = make_solution(TEMPLATE_STR)
    assert solution == T_SOLUTION_STR
    solution = make_solution(FMFT_STR)
    assert solution == FMFS_STR


def test_question_chunks():
    nb = load(SOLUTION_FNAME)
    chunks = question_chunks(nb)
    assert len(chunks) == 15
    nb = loads("""\
Some text

```{python}
# Not question
a = 1
```

More text.

```{r}
# Still not question
b <- 2
```

Another line of text.

```{r}
#- This is a question.
c <- 3
```

Text.

Continues.

```{python}
# This is a question too.
#<- print("hello")
```

Typing is easy but boring.

```{r}
# This is not question, again.
d <- 4
```

```{python}
#- This is a question, again.
e <- 4
```
""")
    chunks = question_chunks(nb)
    assert len(chunks) == 3
    assert [c.code for c in chunks] == [
        '#- This is a question.\nc <- 3\n',
        '# This is a question too.\n#<- print("hello")\n',
        '#- This is a question, again.\ne <- 4\n']
    assert [c.language for c in chunks] == ['r', 'python', 'python']


def test_null_solution():
    # A notebook with no question cells doesn't result in an error.
    nb = RNotebook.from_string('')
    check_marks(nb.nb_str, 0)


def test_check_marks():
    nb = load(SOLUTION_FNAME)
    q_chunks = question_chunks(nb)
    check_chunk_marks(q_chunks)


def test_marks_re():
    assert MARK_RE.match(
        '#- 5 marks / 100 (total 95 so far).').groups() == ('5', '100', '95')


def test_marks():
    assert get_marks('#- 5 marks / 100 (total 95 so far).') == (5, 100, 95)


def test_template2exercise():
    t2e = template2exercise
    assert t2e('#- foo\n#- bar') == '#- foo\n#- bar\n'
    assert t2e('#- foo\na = 1\n#- bar') == '#- foo\n#- bar\n'
    assert t2e('#- foo\na = 1\n# bar') == '#- foo\n'
    assert t2e('#- foo\n#<- a = ?\n# bar') == '#- foo\na = ?\n'
    assert t2e('#- foo\n#<- a = ?\n#- bar') == '#- foo\na = ?\n#- bar\n'
    assert (t2e('#- foo\n  #<- a = ?\n#- bar') ==
            '#- foo\n  a = ?\n#- bar\n')
    with pytest.raises(MarkupError):  # No space after #<-
        t2e('#- foo\n#<-a = ?\n# bar\n')
    # With space suffix, marker adds a blank line to the solution.
    assert t2e('#- foo\n#<- \n# bar') == '#- foo\n\n'
    with pytest.raises(MarkupError):  # No closing #<-
        t2e('#- foo\n#<-\n# bar\n')
    # With a closing marker - include solution code in exercise.
    assert (t2e('#- foo\n#<-\n# bar\na = 1\n#<-\n') ==
            '#- foo\n# bar\na = 1\n')
    # Check stuff after both chunk still gets stripped.
    assert (t2e(
        '#- foo\n#<-\n# bar\na = 1\n#<-\nb = 2\n') ==
        '#- foo\n# bar\na = 1\n')
    # And that one-line #<- still works.
    assert (t2e(
        '#- foo\n#<-\n# bar\na = 1\n#<-\n#<- b = 2\n') ==
        '#- foo\n# bar\na = 1\nb = 2\n')
    # Test a second chunk.
    assert (t2e(
        '#- foo\n#<-\n# bar\na = 1\n#<-\nb = 2\n'
        '#<-\nc = 2\nd=3\n#<-\ne = 4\n') ==
        ('#- foo\n# bar\na = 1\n'
         'c = 2\nd=3\n'))
    # Test both-line
    assert t2e('#- foo\n#<--\na = 1\n#- bar') == '#- foo\na = 1\n#- bar\n'
    assert (t2e('#- foo\n#<--\na = 1\n#- bar\n#<--\n# baz') ==
            '#- foo\na = 1\n#- bar\n# baz\n')
    # Test both-line error
    with pytest.raises(MarkupError):
        t2e('#- foo\n#<--\na = 1\n#- bar\n#<--')
    # Mix both-line and both-section
    assert (t2e(
        '#- foo\n#<--\nq = 99\n#<-\n# bar\na = 1\n#<-\nb = 2\n'
        '#<-\nc = 2\nd=3\n#<-\ne = 4\n') ==
        '#- foo\nq = 99\n# bar\na = 1\nc = 2\nd=3\n')
    # both-line ignored inside both-section
    assert (t2e(
        '#- foo\n#<-\n#<--\n# bar\na = 1\n#<-\nb = 2\n'
        '#<-\nc = 2\nd=3\n#<-\ne = 4\n') ==
        '#- foo\n#<--\n# bar\na = 1\nc = 2\nd=3\n')
    # both-section ignored after both-line
    with pytest.raises(MarkupError):
        t2e('#- foo\n#<--#<-\n# bar\na = 1\n<-\nb = 2\n'
            '#<-\nc = 2\nd=3\n#<-\ne = 4\n')
    assert (t2e(
        '#- foo\n#<--\n#<-\n# bar\na = 1\nb = 2\n'
        '#<-\nc = 2\nd=3\n#<-\ne = 4\n') ==
        '#- foo\n#<-\nc = 2\nd=3\n')


def test_readme_example():
    inp_str = dedent("""
    #- Here you will do a simple assignment.
    #- More description of the assignment.
    #- 5 marks / 100 (total 10 marks so far).
    # This comment gets stripped from the exercise version of the cell.
    # Also this one.  The next line adds the text after #<- to the exercise.
    #<- my_variable = ...
    # This comment and the next code line do not appear in the exercise.
    my_variable = 10
    #<-
    # This comment does appear in the exercise, as well as the following code.
    another_variable = 11
    print("Something")
    #<-
    #<--
    # This line follows the both-line marker, and appears in the exercise.
    # This line does not.
    # Starting at the previous line, we resume normal service.  This and
    # the next line of comments do not appear in the exercise.
    """)
    exp_str = dedent("""
    #- Here you will do a simple assignment.
    #- More description of the assignment.
    #- 5 marks / 100 (total 10 marks so far).
    my_variable = ...
    # This comment does appear in the exercise, as well as the following code.
    another_variable = 11
    print("Something")
    # This line follows the both-line marker, and appears in the exercise.
    """)
    assert template2exercise(inp_str).strip() == exp_str.strip()



""" Utilities for making and checking exercises
"""

import re
from functools import partial
import codecs

from rnbgrader import loads
from rnbgrader.nbparser import Chunk


class ExerciseError(RuntimeError): pass

class MarkError(ExerciseError): pass

class MarkupError(ExerciseError): pass


MARK_RE = re.compile(r"""^\s*\#-
                     \s+([0-9.]+)
                     \s+marks
                     \s+/
                     \s+([0-9.]+)
                     \s+\(total
                     \s+([0-9.]+)""", re.VERBOSE)


EX_COMMENT_RE = re.compile(r'^\s*#<?-', re.M)


def read_utf8(fname):
    with codecs.open(fname, 'r', encoding='utf8') as fobj:
        return fobj.read()


def write_utf8(fname, contents):
    with codecs.open(fname, 'w', encoding='utf8') as fobj:
        return fobj.write(contents)


def question_chunks(nb):
    return [chunk for chunk in nb.chunks if EX_COMMENT_RE.search(chunk.code)]


def get_marks(code):
    for line in code.splitlines():
        match = MARK_RE.match(line)
        if match is not None:
            return tuple(float(v) for v in match.groups())
    return None, None, None


def check_chunk_marks(question_chunks, total=100):
    running = 0
    exp_running = 0
    for chunk in question_chunks:
        msg = (f'chunk:\n\n{chunk.code}\n' +
               f'Chunk starts at line {chunk.start_line}')
        mark, out_of, exp_running = get_marks(chunk.code)
        if mark is None:
            raise MarkError(f'No mark in {msg}')
        if out_of != total:
            raise MarkError(f'Total {out_of} should be {total} in {msg}')
        running += mark
        if running != exp_running:
            raise MarkError(
                f'Running total {running} incorrect; ' +
                f'should be {exp_running} in {msg}')
    if exp_running != total:
        raise MarkError(f'Grand total {exp_running} but should be {total}')


def add_marks(code, total, always=False):
    if not always and get_marks(code)[0] is not None:
        return code
    lines = []
    in_comments = True
    for line in code.splitlines(keepends=True):
        if in_comments and not EX_COMMENT_RE.match(line):
            lines.append('#-  marks / {} (total  so far)\n'.format(total))
            in_comments = False
        lines.append(line)
    return ''.join(lines)


def template2exercise(code):
    """ Convert `code` marked up for exercise + solution to exercise format.

    Parameters
    ----------
    code : str
        String containing one or more lines of code.

    Returns
    -------
    exercise_code : str
        Code as it will appear in the exercise version.
    """
    lines = []
    state = 'default'
    for line in code.splitlines():
        if state == 'both-line':
            lines.append(line)
            state = 'default'
            continue
        sline = line.lstrip()
        is_both_mark = sline == '#<-'  # Start/end of both-mark
        if state == 'both-section':
            if is_both_mark:
                state = 'default'
            else:
                lines.append(line)
            continue
        # Must be default state
        assert state == 'default'
        if is_both_mark:
            state = 'both-section'
            continue
        if sline == '#<--':  # both-line marker - next line ex + solution.
            state = 'both-line'
            continue
        if not sline.startswith('#'):
            continue
        if sline.startswith('#<- '):
            lines.append(line.replace('#<- ', ''))
        elif sline.startswith('#<-'):
            raise MarkupError('There must be a space after the #<- marker '
                              'unless it is a both-line marker or a '
                              'line on its own:\n' + code)
        elif sline.startswith('#-'):
            lines.append(line)
    if state == 'both-line':
        raise MarkupError('No line after #<-- marker:\n' + code)
    if state == 'both-section':
        raise MarkupError('Missing a closing #<- marker:\n' + code)
    return '\n'.join(lines) + '\n'


def template2solution(code):
    """ Convert `code` marked up for exercise + solution to solution format.

    Parameters
    ----------
    code : str
        String containing one or more lines of code.

    Returns
    -------
    solution_code : str
        Code as it will appear in the solution version.
    """
    lines = [L for L in code.splitlines() if not L.strip().startswith('#<-')]
    return '\n'.join(lines) + '\n'


def replace_chunks(nb_str, chunks):
    lines = nb_str.splitlines(keepends=True)
    for chunk in chunks:
        lines[chunk.start_line] = chunk.code
        for line_no in range(chunk.start_line + 1, chunk.end_line + 1):
            lines[line_no] = ''
    return ''.join(lines)


def process_questions(nb, func):
    chunks = question_chunks(nb)
    chunks = [Chunk(func(c.code),
                    c.language,
                    c.start_line,
                    c.end_line)
              for c in chunks]
    return replace_chunks(nb.nb_str, chunks)


def make_filtered(template_str, filt_func):
    return process_questions(loads(template_str), filt_func)


make_exercise = partial(make_filtered, filt_func=template2exercise)

make_solution = partial(make_filtered, filt_func=template2solution)


def check_marks(nb_str, total=100):
    check_chunk_marks(question_chunks(loads(nb_str)), total)


def make_check_exercise(solution_str, total=100):
    exercise = make_exercise(solution_str)
    check_marks(exercise, total)
    return exercise



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Test scripts

Test running scripts
"""

import os
from os.path import abspath

from scripttester import ScriptTester

from rnbgrader.nbparser import read_file

from rnbgrader.tmpdirs import in_dtemp

from . import test_exerciser as te


runner = ScriptTester('rmdex', win_bin_ext='.exe')
run_command = runner.run_command


def script_test(func):
    # Decorator to label test as a script_test
    func.script_test = True
    return func


def assert_wd_fnames(fnames):
    assert sorted(os.listdir('.')) == list(fnames)


@script_test
def test_rmdex():
    script = 'rmdex'
    soln_fname = abspath(te.SOLUTION_FNAME)
    template_fname = abspath(te.TEMPLATE_FNAME)
    with in_dtemp():
        cmd = [script,
               soln_fname,
               'out.Rmd']
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames(['out.Rmd'])
        assert read_file('out.Rmd') == te.EXERCISE_STR
    with in_dtemp():
        # Output filename is optional.
        cmd = [script,
               soln_fname]
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames([])
    with in_dtemp():
        # Solution the same as template for first example.
        cmd = [script,
               soln_fname,
               'out1.Rmd',
               'out2.Rmd']
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames(['out1.Rmd', 'out2.Rmd'])
        assert read_file('out1.Rmd') == te.EXERCISE_STR
        assert read_file('out2.Rmd') == te.SOLUTION_STR
    with in_dtemp():
        # Exercise filename can be underscore - not written.
        cmd = [script,
               soln_fname,
               '_',
               'out3.Rmd']
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames(['out3.Rmd'])
        assert read_file('out3.Rmd') == te.SOLUTION_STR
    with in_dtemp():
        # Solution filename can be _ too (not that it's very useful).
        cmd = [script,
               soln_fname,
               'out4.Rmd',
               '_']
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames(['out4.Rmd'])
        assert read_file('out4.Rmd') == te.EXERCISE_STR
    with in_dtemp():
        # Check marks doesn't need an output filename.
        cmd = [script,
               '--check-marks',
               soln_fname]
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames([])
    with in_dtemp():
        # Second example, just output exercise.
        cmd = [script,
               template_fname,
               'out5.Rmd']
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames(['out5.Rmd'])
        assert read_file('out5.Rmd') == te.T_EXERCISE_STR
    with in_dtemp():
        # Second example, just output solution.
        cmd = [script,
               template_fname,
               '_',
               'out6.Rmd']
        code, stdout, stderr = run_command(cmd)
        assert read_file('out6.Rmd') == te.T_SOLUTION_STR
    with in_dtemp():
        # Second example, output both.
        cmd = [script,
               template_fname,
               'out7.Rmd',
               'out8.Rmd']
        code, stdout, stderr = run_command(cmd)
        assert_wd_fnames(['out7.Rmd', 'out8.Rmd'])
        assert read_file('out7.Rmd') == te.T_EXERCISE_STR
        assert read_file('out8.Rmd') == te.T_SOLUTION_STR



"""Command line interface to RmdEx

* Get code chunks
* Filter code chunks for presence of #- comments, indicating this is a
  question.  Code chunks #<- comments are also questions.
* Comments starting #<- (followed by space) should go into exercise with
  #<- prefix stripped.  Such comments removed in the solution.
* When the whole line is exactly #<- this is a "Both Marker".  It indicates
  that all lines up to the next Both Marker should go in both exercise and
  solution.  Both Markers always removed from exercise and solution.
* Check that each question has marks recorded, when check-marks option
  specified.
* Check that marks add up to given total, when check-marks option
  specified.
* Generate exercise or solution, with modified question chunks.
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter

from rmdex.exerciser import (make_exercise, make_solution, check_marks,
                             read_utf8, write_utf8)


CONVERTERS = {
    'exercise': make_exercise,
    'solution': make_solution,
}


def get_parser():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument("template_rmd", help="filename of template notebook")
    parser.add_argument("exercise_rmd", nargs='?', default='_',
                        help="output filename for exercise notebook; "
                        "default is '_' meaning, omit write of exercise")
    parser.add_argument("solution_rmd", nargs='?', default='_',
                        help="output filename for solution notebook; "
                        "default is '_' meaning, omit write of solution")
    parser.add_argument("--check-marks", action="store_true",
                        help="Enable checking of mark totals")
    parser.add_argument("--total", type=int, default=100,
                        help="Total marks that mark check should sum to")
    return parser


def main_func():
    args = get_parser().parse_args()
    nb_fname = args.template_rmd
    ex_fname = None if args.exercise_rmd == '_' else args.exercise_rmd
    soln_fname = None if args.solution_rmd == '_' else args.solution_rmd
    nb_str = read_utf8(nb_fname)
    # Always create exercise in memory, to check marks.
    if args.check_marks or ex_fname is not None:
        ex_nb = make_exercise(nb_str)
        if args.check_marks:
            check_marks(ex_nb, args.total)
        if ex_fname is not None:
            write_utf8(ex_fname, ex_nb)
    if soln_fname is not None:
        soln_nb = make_solution(nb_str)
        write_utf8(soln_fname, soln_nb)



""" Initialize rmdex package
"""

# Versioneer boilerplate
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


from .exerciser import make_exercise, make_solution, check_marks



#!/usr/bin/env python
# -*- coding: utf-8 -*-
''' Installation script for rmdex package '''
import re
from os.path import join as pjoin

from setuptools import setup

import versioneer

# Get install requirements from requirements.txt file
with open('requirements.txt', 'rt') as fobj:
    install_requires = [line.strip() for line in fobj
                        if line.strip() and not line[0] in '#-']
# Get any extra test requirements
with open('test-requirements.txt', 'rt') as fobj:
    test_requires = [line.strip() for line in fobj
                     if line.strip() and not line[0] in '#-']

# Requires for distutils (only used in pypi interface?)
break_ver = re.compile(r'(\S+?)(\[\S+\])?([=<>!]+\S+)')
requires = [break_ver.sub(r'\1 (\3)', req) for req in install_requires]

cmdclass = versioneer.get_cmdclass()


setup(name='rmdex',
      version=versioneer.get_version(),
      cmdclass=cmdclass,
      description='Utilities to process exercises in RMarkdown notebooks',
      author='Matthew Brett',
      author_email='matthew.brett@gmail.com',
      maintainer='Matthew Brett',
      maintainer_email='matthew.brett@gmail.com',
      url='http://github.com/matthew-brett/rmdex',
      packages=['rmdex',
                'rmdex.tests'],
      package_data = {'rmdex': [
          'tests/data/*',
      ]},
      license='BSD license',
      classifiers = [  # https://pypi.org/classifiers
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Operating System :: MacOS',
        ],
      long_description = open('README.rst', 'rt').read(),
      install_requires = install_requires,
      extras_require = {'test': test_requires},
      # For pip versions >= 9
      python_requires = '>=3.6',
      entry_points={
          'console_scripts': [
              'rmdex = rmdex.cli:main_func',
          ],
      }
      )
